package kohtaushallinta;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.PrintStream;
import java.util.ArrayList;

/**
 * @author Jussi Lipitsäinen
 * @version 6.3.2025
 * Hakee tiedostosta tiedot tietueita yhdistävistä linkeistä. Muodostaa ja poistaa niitä ohjeiden mukaan ja tallentaa ne käskettäessä.
 */
public class Linkit implements TiedostoKayttaja {
    private String tiedosto;
    private Linkki[] linkit;
    
    private static final int KOKO = 200; //oletuskoko
    private static final char KOMMENTTI = '#'; //tiedoston kommenttimerkki

    
    /**
     * Oletusmuodostin
     */
    public Linkit() {
        this.tiedosto = "";
        linkit = new Linkki[KOKO];        
    }
    
    
    /**
     * @param koko Linkit-taulukon koko
     */
    public Linkit(int koko) {
        int k = 0;
        if (koko > 0) k = koko; 
        linkit = new Linkki[k];
    }
    
    
    /**
     * Lataa tiedot annetusta tiedostosta
     * @param koko taulukon koko
     * @param tdst Tiedosto, josta tiedot haetaan     
     */
    public Linkit(int koko, String tdst) {
        this(koko);
        this.tiedosto = tdst;
    }
    
    
    /**
     * Lataa tiedot annetusta tiedostosta
     * @param tdst Tiedosto, josta tiedot haetaan     
     */
    public Linkit(String tdst) {
        this.linkit = new Linkki[KOKO];
        this.tiedosto = tdst;                
    }
    
    
    /**
     * Asettaa tiedoston linkeille
     * @param tiedosto asetettava tiedosto
     */
    public void setTiedosto(String tiedosto) {
        this.tiedosto = tiedosto;
    }
    
    /**
     * Lisää id2:n id1:n linkkeihin. Jos taulukon oliota ei ole alustettu, tekee sen.
     * @param id1 ID, johon linkitetään
     * @param id2 linkitettävä ID
     */
    public void lisaa(int id1, int id2) {
        if (linkit[id1] == null) linkit[id1] = new Linkki();
        linkit[id1].lisaa(id2);
    }
   
    
    /**
     * Jos oliolle määritelty tiedosto on olemassa, lataa siitä tiedot taulukkoon 
     * @throws FileNotFoundException tiedostoa ei löydy
     */
    public void lataa() throws FileNotFoundException {
        //
    }
    
    
    /**
     * Lukee linkit String-matriisista
     * @param tiedot matriisi
     */
    /*
    private void lueMatriisista(String[][] tiedot) {
        for (String[] rivi : tiedot) {
            if (rivi[0].length() < 1 || rivi[0].charAt(0) == KOMMENTTI) continue;
            int a, b;
            try {
                a = Integer.parseInt(rivi[0].strip());
                b = Integer.parseInt(rivi[1].strip());
            } catch (NumberFormatException e) {
                //TODO KÄSITTELE PAREMMIN
                continue;
            }
            this.lisaa(a, b);
        }   
    }
    */
    
    
    /**
     * Palauttaa id:n linkit taulukkona
     * @param id id
     * @return kokonaislukutaulukko, jossa linkitetyt id:t
     */
    public int[] getIDt(int id) {
        if (id < 0 || id >= linkit.length || linkit[id] == null) return new int[] {}; 
        return linkit[id].get();
    }
    
    
    @Override
    public String[] haeRivit() {
        // TODO Auto-generated method stub
        return null;
    }


    @Override
    public void lue(FileInputStream fis) {
        // TODO Auto-generated method stub
        
    }


    @Override
    public void kirjoita(FileWriter fw) {
        // TODO Auto-generated method stub
        
    }
}


class Linkki {
    private ArrayList<Integer> linkitetyt = new ArrayList<Integer>();
    
    
    /**
     * Lisää taulukon tietueet listaan, jos niitä ei siinä jo ole
     * @param uusia taulukko Tietueita
     * @example
     * <pre name="test">
     *  int[] idt = { 3, 5, 6, 5, 7, 1, -1 };
     *  Linkki l = new Linkki();
     *  l.lisaa(idt);
     *  l.toString() === "3 5 6 7 1 ";
     *  idt = new int[] { 3, 5, 2, 7 };
     *  l.lisaa(idt);
     *  l.toString() === "3 5 6 7 1 2 ";
     *  l.lisaa(new int[] {});
     *  l.toString() === "3 5 6 7 1 2 ";
     *  l.lisaa(new int[] { -10 });
     *  l.toString() === "3 5 6 7 1 2 ";
     *  l.lisaa(new int[] {10});
     *  l.toString() === "3 5 6 7 1 2 10 ";
     * </pre>
     */
    public void lisaa(int[] uusia) {
        for (int i : uusia)
            lisaa(i); 
    }
    
    
    /**
     * Poistaa linkeistä id:n, jos se löytyy
     * @param id poistettava id
     */
    public void poista(int id) {
        for(Integer i : linkitetyt)
            if (i.intValue() == id) {
                linkitetyt.remove(i);
                break;
            }
    }
    
    
    /**
     * Hakee id:t taulukkona
     * @return id:t taulukkona
     */
    public int[] get() {
        int[] taulukko = new int[linkitetyt.size()];
        for (int i = 0; i < linkitetyt.size(); i++)
            taulukko[i] = linkitetyt.get(i).intValue(); 
        return taulukko;
    }
    
    
    /**
     * Jos tietuetta ei ole listassa, lisätään se
     * @param id lisättävä ID
     * @example
     * <pre name="test">
     *  Linkki l = new Linkki();
     *  l.etsi(3) === false;
     *  l.lisaa(5);
     *  l.etsi(5) === true;
     *  l.etsi(3) === false;
     * </pre>
     */
    public void lisaa(int id) {
        if (id < 0) return;
        if (this.etsi(id) == false) linkitetyt.add(id);
    }
    
    
    /**
     * Tarkistaa, onko id linkitetty
     * @param id etsittävä id
     * @return true, jos löytyy linkistä. Muuten false
     */
    public boolean etsi(int id) {
        for (Integer i : linkitetyt)
            if (i.intValue() == id) return true;
        return false;
    }
    
    
    @Override
    public String toString() {
        String s = "";
        for (Integer i : linkitetyt) s += i.toString() + " ";
        return s;
    }
}
