package kohtaushallinta;

import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Scanner;

/**
 * @author Jussi Lipitsäinen
 * @version 4.3.2025
 *
 * Käsittelee joukkoa Tietueet-olioita. Järjestää Tietueet niiden id-kentän arvon mukaan,
 * Tietue tallennetaan sen id-kenttää vastaavaan indeksiin. Tarjoaa metodin [kopioi()] sisällön tallentamiseksi
 * uuteen Tietueet-olioon, jonka arvon voi määrittää suuremmaksi.
 */
public class Tietueet implements TiedostoKayttaja {
    private String tiedosto;
    private Tietue[] tietueet;    
    private int lkm;
    private int ekaVapaa; //ensimmäinen vapaa indeksi    
    
    /** OLETUSARVOT */
    private static final char KOMMENTTI = '#';  //Kommenttirivin merkki tiedostossa
    private static final char EROTIN = '|';     //Kenttiä erottava merkki tiedostossa
    private static final int MAXLKM = 200;      //Tieteueiden maksimimäärä
    
    
    /**
     * Luo Tietueet, jossa on tilaa 200 tietueelle
     */
    public Tietueet() {
        this.tietueet = new Tietue[MAXLKM];        
        this.lkm = 0;
        this.ekaVapaa = 0;
        this.tiedosto = "";
    }
    
    
    /**
     * Luo Tietueet 200 tietueelle ja lataa tiedot annetusta tiedostosta
     * @param tdst tallennustiedosto
     */
    public Tietueet(String tdst) {
        this.tietueet = new Tietue[MAXLKM];
        this.tiedosto = tdst;
        this.lkm = 0;
        this.ekaVapaa = 0;
    }
    
    
    /**
     * Lukee tiedot tiedostosta
     * @param maxLkm tietueiden maksimimäärä
     * @param tdst tallennustiedosto
     */
    public Tietueet(int maxLkm, String tdst) {
        this.tietueet = new Tietue[maxLkm];
        this.tiedosto = tdst;
        this.lkm = 0;
        this.ekaVapaa = 0;
    }
   
    
    /**
     * Hakee Tietueiden käyttämän tiedoston nimen
     * @return tiedostonimen
     */
    public String getTiedosto() {
        return this.tiedosto;
    }
    
    
    /**
     * Tarkistaa, montako Tietuetta on tallennettuna
     * @return tallennettujen Tietueiden lukumäärä
     */
    public int getLkm() {
        return this.lkm;
    }
    
    
    /**
     * Tarkistaa Tietueiden maksimilukumäärän
     * @return maksimilukumäärän
     */
    public int getMaxLkm() {
        return this.tietueet.length;
    }
    
    
    /**
     * Hakee tietueen ID-numeron perusteella
     * @param id tietueen ID
     * @return Tietueen tai null, jos ID-numerolla ei löydy tietuetta
     * @throws IndexOutOfBoundsException ID on tietuetaulukon indeksiarvojen ulkopuolella
     * Tietueet tt = new Tietueet(10, hahmot);
     * tt.get(0).getID() === 0;
     * tt.get(-5) #throws IndexOutOfBoundsException;
     */
    public Tietue get(int id) throws IndexOutOfBoundsException {
        if (id < 0 || id >= this.getMaxLkm()) throw new IndexOutOfBoundsException("Sallitut ID-arvot: 0-" + (this.getMaxLkm()-1) +
                                                                                                ". ID " + id + " on välin ulkopuolella.");
        return this.tietueet[id];
    }
  
    
    /**
     * Kopioi tiedot uuteen olioon, jonka koko on annettu. Uuden koon on oltava vähintään yhtä suuri kuin oma. Ei tee syväkopiota.
     * @param maxLkm uuden olion koko
     * @return uuden Tietueet, johon on tallennettu omat tiedot. Jos maxLkm on liian pieni, null.
     * @example
     * <pre name="test">
     *  Tietueet tt1 = new Tietueet(20, "");
     *  tt1.lisaa(new Tietue(1, "Tietue1"));
     *  tt1.lisaa(new Tietue(10, "Tietue2"));
     *  tt1.lisaa(new Tietue(11, "Tietue3"));
     *  tt1.lisaa(new Tietue(5, "Tietue4"));
     *  Tietueet tt2 = tt1.kopioi(10);
     *  tt2 === null;
     *  tt2 = tt1.kopioi(30);  
     *  tt1.get(1).equals(tt2.get(1)) === true;
     *  tt1.get(10).equals(tt2.get(10)) === true;
     *  tt1.get(11).equals(tt2.get(11)) === true;
     *  tt1.get(5).equals(tt2.get(5)) === true;
     * </pre>
     */
    public Tietueet kopioi(int maxLkm) {
        if (maxLkm < this.getMaxLkm()) return null;
        Tietueet uusi = new Tietueet(maxLkm, this.tiedosto);        
        for (Tietue t : this.tietueet)
            if (t != null) uusi.lisaa(t);
        return uusi;
    }
        
    
    /**
     * Asettaa tiedostonimen
     * @param tdst tiedosto
     */
    public void setTiedosto(String tdst) {
        this.tiedosto = tdst;        
    }
    
    
    /**
     * Tarkistaa, onko tilaa uudelle tietueelle
     * @return true, jos on tilaa.
     */
    public boolean onTilaa() {
        return (this.ekaVapaa != -1);
    }
    
    
    /**
     * Tarkistaa, missä on ensimmäinen vapaa tila
     * @return vapaan paikan indeksin. Jos tilaa ei ole, palauttaa -1.
     */
    public int ekaVapaa() {
        return this.ekaVapaa;
    }
    
    
    /**
     * Tarkistaa, missä kohdassa on ensimmäinen vapaa tila.
     * Jos saavutetaan taulukon loppu, tarkistetaan taulukko.
     */
    private void tarkistaVapaa() {
        for (int i = this.ekaVapaa; i < this.tietueet.length; i++)
            if (tietueet[i] == null) {
                this.ekaVapaa = i;                
                return;
            }
        // Taulu on ilmeisesti täysi. Tarkistetaan vielä varmuuden vuoksi tilanne ja päivitetään muuttujat
        boolean loytyi = false;
        int t = 0;
        for (int i = 0; i < this.ekaVapaa; i++) {
            if (tietueet[i] != null) t++;
            else if (!loytyi) {
                loytyi = true;
                this.ekaVapaa = i;
            }            
        }
        if (!loytyi) this.ekaVapaa = -1;
        this.lkm = t;        
    }
    
    
    /**
     * Lisää tietueen listaan oman id-numeronsa ilmoittamaan paikkaan.
     * @param tietue lisättävä tietue
     * @example
     * <pre name="test">
     *  Tietueet tt = new Tietueet(20, "hahmot");
     *  tt.lisaa(new Tietue(1, "Otsikko"));
     *  tt.get(1).getOtsikko() === "Otsikko";
     *  try {
     *      tt.lisaa(new Tietue(1, "ToinenOtsikko"));
     *  } catch (IllegalArgumentException ie) {
     *      ie.getMessage() === "Tietue ID:llä 1 on jo olemassa."
     *  }
     *  tt.get(1).getOtsikko() === "Otsikko";
     *  try {
     *      tt.lisaa(new Tietue(25, "Otsikkoo"));
     *  } catch (IndexOutOfBoundsException ie) {
     *      ie.toString() === "java.lang.IndexOutOfBoundsException: Sallitut ID-arvot: 0-19. ID 25 on välin ulkopuolella.";
     *  }
     *  try {
     *      tt.lisaa(new Tietue(-1, "Otsikko2", new Avainsanat(10, "Koe, Koe2"), "Kuvaus"));
     *  } catch (IndexOutOfBoundsException ie) {
     *      ie.toString() === "java.lang.IndexOutOfBoundsException: Sallitut ID-arvot: 0-19. ID -1 on välin ulkopuolella.";
     *  }
     *  tt.lisaa(new Tietue(0, "Otsikko2", new Avainsanat(10, "Koe, Koe2"), "Kuvaus"));
     *  tt.get(0).getAvainsanat().get(1) === "Koe2";
     *  tt.get(0).getKuvaus() === "Kuvaus";
     *  try {
     *      tt.lisaa(new Tietue(5, "Otsikko"));
     *  } catch (IllegalArgumentException ie) {
     *      ie.getMessage() === "Tietue otsikolla \"Otsikko\" on jo olemassa."
     *  }
     * </pre>
     */
    public void lisaa(Tietue tietue) {        
        if (!this.onTilaa()) throw new IllegalStateException("Tietueissa ei ole tilaa lisätä uutta tietuetta.");
        if (this.etsi(tietue.getOtsikko()) > -1) throw new IllegalArgumentException("Tietue otsikolla \"" + tietue.getOtsikko() + "\" on jo olemassa."); 
        int i = tietue.getID();
        if (i >= this.tietueet.length || i < 0) throw new IndexOutOfBoundsException("Sallitut ID-arvot: 0-" + (this.getMaxLkm()-1) +
                                                                                    ". ID " + i + " on välin ulkopuolella.");
        if (this.tietueet[i] != null) throw new IllegalArgumentException("Tietue ID:llä " + tietue.getID() + " on jo olemassa.");
        
        this.tietueet[i] = tietue;
        this.lkm++;
        if (this.ekaVapaa == i) {
            this.tarkistaVapaa();
        }        
    }
    
    
    /**Poistaa Tietueen id:n perusteella.
     * @param id poistettava Tietue
     * @throws IndexOutOfBoundsException ID on taulukon rajojen ulkopuolella
     * @throws IllegalArgumentException ID viittaa tietueeseen, jota ei ole olemassa
     */
    public void poista(int id) throws IndexOutOfBoundsException, IllegalArgumentException {
        if (id < 0 || id >= this.tietueet.length) throw new IndexOutOfBoundsException("Sallitut ID-arvot: 0-" + (this.getMaxLkm()-1) +
                                                                                        ". Yritettiin lisätä tietue ID:llä "+id);
        Tietue t = this.get(id);
        if (t == null) throw new IllegalArgumentException("Yritettiin poistaa tietuetta, jota ei ole olemassa: ID "+ id);
        t = null;        
        this.lkm--;
        if (this.ekaVapaa == -1 || this.ekaVapaa > id) this.ekaVapaa = id;
    }
    
    
    /**
     * Hakee tallennettujen tietueiden id:t taulukkona
     * @return taulukon id-numeroista
     * 
     */
    public Tietue[] getTietueet() {
        Tietue[] tt = new Tietue[lkm];
        int i = 0;
        for (Tietue t : this.tietueet)
            if (t!= null) tt[i++] = t;
        return tt;
    }
     
    
    /**Etsii tietueen indeksin, jolla on sama otsikko.
     * @param otsikko haettava otsikko
     * @return tietueen indeksin. Jos ei löydy, -1.
     * @example
     * <pre name="test">
     *  Tietueet tt = new Tietueet (10, "hahmot.dat");
     * </pre>
     */
    public int etsi(String otsikko) {
        for (int i = 0; i < this.tietueet.length; i++)
            if (tietueet[i] != null && tietueet[i].getOtsikko().equals(otsikko)) return i;
        return -1;
    }
    
    
    /**
     * Jos on tilaa, luo uuden Tietueen annetuilla parametreillä. Palauttaa luodun tietueen.
     * Jos ei ole tilaa tai otsikko on jo käytössä, palautetaan null.
     * @param otsikko otsikko
     * @param avainsanat avainsanat
     * @param kuvaus kuvaus
     * @return luodun Tietueen. Jos luominen ei onnistu, null.
     * @example
     * <pre name="test">
     *  Tietueet tt = new Tietueet(2, "hahmot");
     *  tt.onTilaa() === true; 
     *  tt.luoTietue ("Eka", new String[] {"Eka", "Toka"}, "");
     *  tt.get(0).getOtsikko() === "Eka";
     *  tt.luoTietue ("Eka", new String[] {"Kolmas"}, "") === null;
     *  tt.onTilaa() === true;
     *  tt.get(0).getAvainsanat().toString() === "Eka, Toka";
     *  tt.get(0).getOtsikko() === "Eka";  
     *  tt.luoTietue ("Toka", new String[] {"Eka", "Toka"}, "");
     *  tt.onTilaa() === false;
     *  tt.get(0).getOtsikko() === "Eka";
     *  tt.get(1).getOtsikko() === "Toka"; 
     *  tt.luoTietue ("Kolmas", new String[] {"Eka", "Toka"}, "") === null;
     *  tt.get(0).getOtsikko() === "Eka";
     *  tt.get(1).getOtsikko() === "Toka"
     * </pre>
     */
    public Tietue luoTietue(String otsikko, String[] avainsanat, String kuvaus) {
        if (this.ekaVapaa == -1 || this.etsi(otsikko) != -1) return null;
        int id = this.ekaVapaa;
        this.tietueet[id] = new Tietue(id, otsikko, new Avainsanat(avainsanat), kuvaus);
        this.tarkistaVapaa();
        return this.tietueet[id];
    }
    
    
    /**
     * Luo uuden Tietueen, jossa on annettu otsikko ja palauttaa sen.
     * Jos ei ole tilaa tai otsikko on jo käytössä, palautetaan null.
     * @param otsikko otsikko
     * @return uuden Tietueen  tai null, jos luominen ei onnistu
     * @example
     * <pre name="test">
     *  Tietueet tt = new Tietueet(2, "hahmot");
     *  tt.luoTietue("Eka").getID() === 0;
     *  tt.luoTietue("Eka") === null;
     *  tt.luoTietue("Toka").getID() === 1;
     * </pre>
     */
    public Tietue luoTietue(String otsikko) {
        return this.luoTietue(otsikko, new String[0], "");
    }
    
    
    /**
     * Luo uuden Tietueen otsikolla "Nimi" ja jos on varattu, "NimiX", jossa X = ensimmäinen vapaa kokonaisluku väliltä 1-MAX_Value 
     * @return uuden Tietueen
     * @example
     * <pre name="test">
     *  Tietueet tt = new Tietueet("koe.dat");
     *  tt.luoTietue();
     *  tt.get(0).getOtsikko() === "Nimi";
     *  tt.get(1) === null;
     *  tt.luoTietue();
     *  tt.luoTietue();
     *  tt.get(1).getOtsikko() === "Nimi1";
     *  tt.get(2).getOtsikko() === "Nimi2";
     * </pre>
     */
    public Tietue luoTietue() {
        String otsikko = "Nimi";
        if (this.etsi(otsikko) == -1) return luoTietue(otsikko);
        for (int i = 1; i < Integer.MAX_VALUE; i++) {
            String vo = otsikko + i;
            if (this.etsi(vo) == -1) return luoTietue(vo);
        }
        return null;
    }
    
    
    /**
     * Tallentaa Tietueet tiedostoon
     * @throws IOException Tiedostoon kirjoittaminen ei onnistu
     */
    public void tallenna() throws IOException {
        Tiedostokasittely.tallennaTiedostoon(this, this.tiedosto);
    }
    
    
    /**
     * Lataa tietueen tiedot tiedostosta
     * @throws IOException Tiedostoa ei löydy tai sitä ei voinut lukea
     */
    public void lataa() throws IOException {        
        Tiedostokasittely.lueTiedostosta(this, this.getTiedosto());
    }
    
    
    @Override
    public String[] haeRivit() {
        String[] rivit = new String[lkm];
        int i = 0;
        for (Tietue t : this.tietueet)
            if (t != null) rivit[i++] = t.toString(); 
        return null;
    }
 

    private void lueRivi(String rivi) {
        if (rivi.length() == 0 || rivi.charAt(0) == KOMMENTTI) return;
        String[] palat = rivi.split("[" + EROTIN + "]");
        try {
            Tietue t = Tietue.lueTaulukko(palat);
            if (t != null) this.lisaa(t);
        } catch (Exception e) {
            System.err.println(e);
        }
    }


    @Override
    public void lue(FileInputStream fis) {
        try (Scanner skanneri = new Scanner(fis)) {
            this.tietueet = new Tietue[tietueet.length];
            while (skanneri.hasNext())
                this.lueRivi(skanneri.nextLine());
        }
    }


    @Override
    public void kirjoita(FileWriter fw) {
        // TODO Auto-generated method stub
        
    }
}
